import java.awt.*;
import java.applet.*;
import java.awt.event.*;
import java.lang.*; 
import java.net.*; 


/**
 * Klasse ComplexViewer (Applet)
 * Dieses Applet liefert eine Darstellung verschiedener komplexer Funktionen und ihrer Ableitung 
 *
 * @author Keith Orpen, Math department, University of British Columbia.
 * @author Porting, interface, and extensions by Djun Kim, Mathematics UBC.
 * @author extended by Bernd Funovits (1) 
 * @author extended by Christoph Schleifer (1) 
 * @author extended by Heinrich Gudenus (2) 
 * @version November, 1995 - First version -  Keith Orpen
 * @version December, 1995 - Ported to Beta API, cosine and Exp functions added, user interface improved. - Djun Kim
 * @version March, 1996 - Added Double buffering, cleaned up code. - Djun Kim
 * @version (1) Erweiterung der Anzahl der Funktionen und Einbindung eines Zooms
 * @version (2) 2006, Erweiterung durch die Darstellung der komplexen Ableitung
 * 
 * Copyright (c)1995 Keith Orpen.  Extensions copyright (c)1995 Djun Kim.
 * The authors reserve all rights.        
 */
public class ComplexViewer extends Applet {
   
    /** Diese Variable wird fuer double buffering verwendet */
    public Image hiddenimagebuffer;
    /** Diese Variable definiert das Graphics-Object, welches auf das Image hiddenimagebuffer angelegt wird. */
    public Graphics hiddengraphics;

    /** Diese Methode legt das Layout von ComplexViewer fest und initialisiert die Variablen hiddenimagebuffer und hiddengraphics. */
    public void init() {
        setLayout(new BorderLayout());
        ViewPanel vp = new ViewPanel(this);
        add("Center",vp);
        add("East", new Controls(vp, vp.getSize()){
                public Dimension getPreferredSize(){
                    return new Dimension(100,300); 
                }
            });
        // fuer das Double-buffering:
        hiddenimagebuffer = createImage(this.getSize().width, this.getSize().height);
        hiddengraphics = hiddenimagebuffer.getGraphics();
        
        repaint();
    }
    
    /** Diese Methode wird wirksam, falls ComplexViewer als Frame und nicht als Applet ausgefuehrt wird. */
    public boolean handleEvent(Event e) {
        switch (e.id) {
        case Event.WINDOW_DESTROY:
            System.exit(0);
            return true;
        default:
            return false;
        }
    }

    /** Diese Methode wird wirksam, falls ComplexViewer als Frame und nicht als Applet ausgefuehrt wird. */       
    public static void main(String args[]) {
        Frame f = new Frame("ComplexViewer");
        ComplexViewer view = new ComplexViewer();
        view.init();
        view.start();
        f.add("Center", view);
        f.show();
    }    
}

/** Diese Klasse definiert den Bereich der graphischen Ausgabe */
class ViewPanel extends Panel {
    
    /** Diese Variable verweist auf das Applet ComplexViewer */
    ComplexViewer parent;

    /** Diese Variable weist der Funktion "Sin" eine Zahl zur Verarbeitung zu */    
    public static final int Sin = 0;
    /** Diese Variable weist der Funktion "Cos" eine Zahl zur Verarbeitung zu */ 
    public static final int Cos = 1;
    /** Diese Variable weist der Funktion "Exp" eine Zahl zur Verarbeitung zu */ 
    public static final int Exp = 2;
    /** Diese Variable weist der Funktion "z_Quadrat" eine Zahl zur Verarbeitung zu */ 
    public static final int z_Quadrat = 3;
    /** Diese Variable weist der Funktion "Sqrt" eine Zahl zur Verarbeitung zu */ 
    public static final int Sqrt = 4;
    /** Diese Variable weist der Funktion "Log" eine Zahl zur Verarbeitung zu */ 
    public static final int Log = 5;
    /** Diese Variable weist der Funktion "Cosh" eine Zahl zur Verarbeitung zu */ 
    public static final int Cosh = 6;
    /** Diese Variable weist der Funktion "Sinh" eine Zahl zur Verarbeitung zu */ 
    public static final int Sinh = 7;

    /** Diese Variable speichert die aktuell ausgewhlte Funktion */
    private int function = Sin;

    /** Dieses Feld dient der Beschreibung der Funktionen: Namen der Funktionen */
    public static final String[] function_names = {"Sin","Cos","Exp","z zum Quadrat","Quadratwurzel","Log","Cosh","Sinh"};
    /** Dieses Feld dient der Beschreibug der Funktionen: Anzahl der Graphen der einzelnen Funktionen */
    public static final int[] CoG_perFunction = {1,1,1,1,2,1,1,1};

    /** Diese Variable definiert den Zoom-Faktor */    
    public static float scale = 0.05f;

    /** Diese Variable definiert den "Eingabe"-Raster (1) */
    private static int gridspace = 4;
    /** Diese Variable definiert den "Eingabe"-Raster (2) */
    private static int gridlines = 2;
    /** Diese Variable definiert den "Eingabe"-Raster (3) */
    private static int gridsize = 2 * gridlines +1;

    /** Diese Variable definiert die X-Koordinate des Mittelpunktes des Eingabe Rasters */
    private int x1 = 250; 
    /** Diese Variable definiert die Y-Koordinate des Mittelpunktes des Eingabe Rasters */
    private int y1 = 150;

    /** Diese Variable definiert die X-Koordinate des Panel-Ursprungs */    
    public static int origin_x = 250;
    /** Diese Variable definiert die Y-Koordinate des Panel-Ursprungs */
    public static int origin_y = 150;
    
    /** Diese Variable definiert die X-Koordinate von z0 (fuer die komplexe Ableitung) */
    public int init_z0_re;
    /** Diese Variable definiert die Y-Koordinate von z0 (fuer die komplexe Ableitung) */
    public int init_z0_im;
    /** Diese Variable definiert die X-Koordinate von z (fuer die komplexe Ableitung) */
    public int init_z_re;
    /** Diese Variable definiert die Y-Koordinate von z (fuer die komplexe Ableitung) */
    public int init_z_im;

    /** Diese Variable beschreibt den Status der Checkbox "derivative" */
    public static boolean cd_state = false;
    /** Diese Variable beschreibt den Status des Buttons "choose z0" */
    public static boolean z0_state = false;
    /** Diese Variable beschreibt den Status der Checkbox "show f(z0)" */
    public static boolean shw1_state = false;
    /** Diese Variable beschreibt den Status der Checkbox "show df(z0)" */
    public static boolean shw2_state = false;
    /** Diese Variable beschreibt den Status der Checkbox "show ap:f(z)" */
    public static boolean shw3_state = false;
    /** Diese Variable beschreibt den Status der Checkbox "show f(z)" */
    public static boolean shw4_state = false;

    /** 
     * Dies ist der Konstruktor fuer Viewpanel.
     *
     * @param target die Instanz des aufrufenden ComplexViewer-Objektes
    */
    ViewPanel(ComplexViewer target) {
        this.parent = target;
    }
    
    /** java.applet.Applet method */
    public void init() {
        setSize(400,400);
    }
    /** java.applet.Applet method */
    public void start() {
    }
    /** java.applet.Applet method */
    public void stop() {
    }
    /** java.applet.Applet method */
    public void destroy() {
    }
    

    /**
     * Mithilfe dieser Methode wird eine gewhlte Funktion als aktuelle Funktion gesetzt
     *  
     * @param function Der Integer-Wert der Funktion, die gesetzt werden soll (sin=0, cos=1, ... , cosh=6, sinh=7) 
     */
    public void setFunction(int function) {
        switch (function) {
        case Sin: 
        case Cos:
        case Exp:
        case z_Quadrat:
        case Sqrt:
        case Log:
        case Cosh:
        case Sinh:
            this.function = function;
            break;
        default:
            throw new IllegalArgumentException();
        } 
    }

    /**
     * Diese Methode dient der Konvertierung von Panel-Koordinaten in Gauss-Koordinaten
     *
     * @param panel_x Real-Teil der komplexen Zahl in Panel-Koordinaten  
     * @return Real-Teil der komplexen Zahl in Gauss-Koordinaten
     */
    private double X_PanelToGauss ( double panel_x ){
        return ( ( panel_x - getSize().width/2 ) * scale );
    }

    /**
     * Diese Methode dient der Konvertierung von Panel-Koordinaten zu Gauss-Koordinaten
     *
     * @param panel_y Imaginaer-Teil der komplexen Zahl in Panel-Koordinaten  
     * @return Imaginaer-Teil der komplexen Zahl in Gauss-Koordinaten
     */
    private double Y_PanelToGauss ( double panel_y ){
        return ( ( panel_y - getSize().height/2 ) * scale ); 
    }

    /**
     * Diese Methode dient der Konvertierung von Gauss-Koordinaten in Panel-Koordinaten
     *
     * @param gauss_x Real-Teil der komplexen Zahl in Gauss-Koordinaten  
     * @return Real-Teil der komplexen Zahl in Panel-Koordinaten
     */
    private long X_GaussToPanel ( double gauss_x ){
        return ( Math.round( gauss_x / scale + getSize().width/2 ) );
    }

    /**
     * Diese Methode dient der Konvertierung von Gauss-Koordinaten in Panel-Koordinaten
     *
     * @param gauss_y Imaginaer-Teil der komplexen Zahl in Gauss-Koordinaten  
     * @return Imaginaer-Teil der komplexen Zahl in Panel-Koordinaten
     */
    private long Y_GaussToPanel ( double gauss_y ){
        return ( Math.round( ( gauss_y / scale + getSize().height/2 ) ) ); 
    }  
  
    /**
     * Diese Methode dient der Konvertierung von Kartesischen-Koordinaten in Polar-Koordinaten, 
     *
     * @param re Real-Teil der komplexen Zahl 
     * @param im Imaginaer-Teil der komplexen Zahl 
     * @return Ein zwei-elementiges Feld, das die komplexe Zahl in ihrer Polardarstellung (r,phi) beinhaltet
     */    
    private double[] CartesianToPolar (double re, double im) {
        double[] r_phi = new double[2];

        r_phi[0] = Math.sqrt(re*re + im*im);
        r_phi[1] = Math.atan2(im,re);

        return r_phi;
    }

    /**
     * Diese Methode dient der Konvertierung von Polar-Koordinaten in Kartesischen-Koordinaten, 
     *
     * @param r Radius 
     * @param phi Argument 
     * @return Ein zwei-elementiges Feld, das die komplexe Zahl in ihrer Kartesischendarstellung (re,im) beinhaltet
     */  
    private double[] PolarToCartesian (double r, double phi) {
        double[] re_im = new double[2];

        re_im[0] = r * Math.cos(phi);
        re_im[1] = r * Math.sin(phi);

        return re_im;
    }

    /*
     * Diese Methode stellt eine einfache Signum Funktion dar
     *
     * @param d Zahl, von der das Vorzeichen zu bestimmen ist
     * @return Vorzeichen (1,0,-1) der uebergebenen Zahl
     */
    private int signum(double d) {
        if ( d < 0 )
            return -1;
        else if ( d == 0 )
            return 0;
        else 
            return 1;
    }

    /*
     * Diese Methode dient der Kalkulation des Real-Teils der Funktionswerte
     *
     * @param re Real-Teil der komplexen Zahl in Gauss-Koordinaten
     * @param im Imaginaer-Teil der komplexen Zahl in Gauss-Koordinaten
     * @return zwei-elementiges Feld, das den Funktionswert des Real-Teils beinhaltet 
     * (falls die Funktion "sqrt" aktiviert ist, werden an dieser Stelle zwei Funktionswerte ausgegeben)
     */    
    private double[] f_re( double re, double im ) {

        double[] ret = new double[1];

        switch (function) {
        case Sin: 
            ret[0] = Math.sin( re ) * ( Math.exp( im ) + Math.exp( -im ) ) / 2;
            return ret;
        case Sinh:
            ret[0] = ( Math.exp( re ) * Math.cos( im ) - Math.exp( -re ) * Math.cos( -im ) ) / 2;
            return ret;
        case Cos:
            ret[0] = Math.cos( re ) * ( Math.exp( im ) + Math.exp( -im ) ) / 2;
            return ret;
        case Cosh:
            ret[0] = ( Math.exp( re ) * Math.cos( im ) + Math.exp( -re ) * Math.cos( -im ) ) / 2;
            return ret;
        case Exp:
            ret[0] = Math.exp( re ) * Math.cos( im );
            return ret;
        case Log:
            ret[0] = Math.log(CartesianToPolar( re, im )[0]);
            return ret;
        case z_Quadrat:
            ret[0] = PolarToCartesian(Math.pow(CartesianToPolar(re,im)[0],2), 2*CartesianToPolar(re,im)[1] )[0];
            return ret;
        case Sqrt:
            ret = new double[2];
            for (int i = 0; i < 2; i++)
                ret[i] = PolarToCartesian(Math.sqrt(CartesianToPolar(re,im)[0]) , CartesianToPolar(re,im)[1]/2 + i*Math.PI )[0];
            return ret;
        default:
            throw new IllegalArgumentException();
        }
    }

    /**
     * Diese Methode dient der Kalkulation des Imaginaer-Teils der Funktionswerte
     *
     * @param re Real-Teil der komplexen Zahl in Gauss-Koordinaten
     * @param im Imaginaer-Teil der komplexen Zahl in Gauss-Koordinaten
     * @return zwei-elementiges Feld, das den Funktionswert des Imaginaer-Teils beinhaltet 
     * (falls die Funktion "sqrt" aktiviert ist, werden an dieser Stelle zwei Funktionswerte ausgegeben) 
    */
    private double[] f_im( double re, double im ) {
        
        double[] ret = new double [1];
        
        switch (function) {
        case Sin:
            ret[0] = Math.cos( re ) * ( Math.exp( im ) - Math.exp( -im ) ) / 2; 
            return ret;
        case Sinh:
            ret[0] = ( Math.exp( re ) * Math.sin( im ) - Math.exp( -re ) * Math.sin( -im ) ) / 2;
            return ret;
        case Cos:
            ret[0] = Math.sin( re ) * ( Math.exp( -im ) - Math.exp( im ) ) / 2;
            return ret;
        case Cosh:
            ret[0] = ( Math.exp( re ) * Math.sin( im ) + Math.exp( -re ) * Math.sin( -im ) ) / 2;
            return ret;
        case Exp:
            ret[0] = Math.exp( re ) * Math.sin( im );
            return ret;
        case Log:
            ret[0] = CartesianToPolar( re, im )[1];
            return ret;
        case z_Quadrat:
            ret[0] = PolarToCartesian(Math.pow(CartesianToPolar(re,im)[0],2), 2*CartesianToPolar(re,im)[1] )[1];
            return ret;
        case Sqrt:
            ret = new double[2];
            for (int i = 0; i < 2; i++)
                ret[i] = PolarToCartesian(Math.sqrt(CartesianToPolar(re,im)[0]) , CartesianToPolar(re,im)[1]/2 + i*Math.PI )[1];
            return ret;
        default:
            throw new IllegalArgumentException();
        }      
    }
         
    /**
     * Diese Methode dient der Kalkulation des Real-Teils der Ableitungsfunktion
     *
     * @param re Real-Teil der komplexen Zahl in Gauss-Koordinaten
     * @param im Imaginaer-Teil der komplexen Zahl in Gauss-Koordinaten
     * @return zwei-elementiges Feld, das den Real-Teil der Ableitungsfunktion beinhaltet 
     * (falls die Funktion "sqrt" aktiviert ist, werden an dieser Stelle zwei Funktionswerte ausgegeben) 
     */ 
    private double[] df_re( double re, double im ) {

        double[] ret = new double[1];
        double rez, imz;

        switch (function) {
        case Sin: 
            ret[0] = Math.cos( re ) * ( Math.exp( im ) + Math.exp( -im ) ) / 2;
            return ret;
        case Sinh:
            ret[0] = ( Math.exp( re ) * Math.cos( im ) + Math.exp( -re ) * Math.cos( -im ) ) / 2;
            return ret;
        case Cos:
            ret[0] = (-1) * Math.sin( re ) * ( Math.exp( im ) + Math.exp( -im ) ) / 2;
            return ret;
        case Cosh:
            ret[0] = ( Math.exp( re ) * Math.cos( im ) - Math.exp( -re ) * Math.cos( -im ) ) / 2;
            return ret;
        case Exp:
            ret[0] = Math.exp( re ) * Math.cos( im );
            return ret;
        case Log:
            ret[0] = re / ( re * re + im * im );
            return ret;
        case z_Quadrat:
            ret[0] = 2 * re; 
            return ret;
        case Sqrt:
            ret = new double[2];
            for (int i = 0; i < 2; i++) {
                rez = PolarToCartesian(Math.sqrt(CartesianToPolar(re,im)[0]) , CartesianToPolar(re,im)[1]/2 + i*Math.PI )[0];
                imz = PolarToCartesian(Math.sqrt(CartesianToPolar(re,im)[0]) , CartesianToPolar(re,im)[1]/2 + i*Math.PI )[1];;
                ret[i] = 2 * ( rez / ( rez * rez + imz * imz ) );
            }            
            return ret;
        default:
            throw new IllegalArgumentException();
        }
    }

    /**
     * Diese Methode dient der Kalkulation des Imaginaer-Teils der Ableitungsfunktion
     *
     * @param re Real-Teil der komplexen Zahl in Gauss-Koordinaten
     * @param im Imaginaer-Teil der komplexen Zahl in Gauss-Koordinaten
     * @return zwei-elementiges Feld, das den Imaginaer-Teil der Ableitungsfunktion beinhaltet 
     * (falls die Funktion "sqrt" aktiviert ist, werden an dieser Stelle zwei Funktionswerte ausgegeben) 
     */   
    private double[] df_im( double re, double im ) {
        
        double[] ret = new double [1];
        double rez,imz;
        
        switch (function) {
        case Sin:
            ret[0] =  Math.sin( re ) * ( Math.exp( -im ) - Math.exp( im ) ) / 2; 
            return ret;
        case Sinh:
            ret[0] = ( Math.exp( re ) * Math.sin( im ) + Math.exp( -re ) * Math.sin( -im ) ) / 2;
            return ret;
        case Cos:
            ret[0] =  (-1) * Math.cos( re ) * ( Math.exp( im ) - Math.exp( -im ) ) / 2;
            return ret;
        case Cosh:
            ret[0] = ( Math.exp( re ) * Math.sin( im ) - Math.exp( -re ) * Math.sin( -im ) ) / 2;
            return ret;
        case Exp:
            ret[0] = Math.exp( re ) * Math.sin( im );
            return ret;
        case Log:
            ret[0] = im / (re * re + im * im);
            return ret;
        case z_Quadrat:
            ret[0] = 2 * im; 
            return ret;
        case Sqrt:
            ret = new double[2];
            for (int i = 0; i < 2; i++) {
                rez = PolarToCartesian(Math.sqrt(CartesianToPolar(re,im)[0]) , CartesianToPolar(re,im)[1]/2 + i*Math.PI )[0];
                imz = PolarToCartesian(Math.sqrt(CartesianToPolar(re,im)[0]) , CartesianToPolar(re,im)[1]/2 + i*Math.PI )[1];
                ret[i] = 2 * ( imz / ( rez * rez + imz * imz ) ) ;
            }
            return ret;
        default:
            throw new IllegalArgumentException();
        }      
    }    

    /**
     * Diese Methode berechnet den Real-Teil des Funktionswertes
     *
     * @param x Real-Teil der komplexen Zahl in Panel-Koordinaten 
     * @param y Imaginaer-Teil der komplexen Zahl in Panel-Koordinaten
     * @param t Nummer des Graphen ("sqrt": 2, ansonsten 1)
     * @return Real-Teil des Funktionswertes in Gauss-Koordinaten
     */
    private long FValue_x(int x, int y, int t) {
        return (X_GaussToPanel(f_re(X_PanelToGauss(x), Y_PanelToGauss(y))[t]));
    }

    /**
     * Diese Methode berechnet den Imaginaer-Teil des Funktionswertes
     *
     * @param x Real-Teil der komplexen Zahl in Panel-Koordinaten 
     * @param y Imaginaer-Teil der komplexen Zahl in Panel-Koordinaten
     * @param t Nummer des Graphen ("sqrt": 2, ansonsten 1)
     * @return Imaginaer-Teil des Funktionswertes in Gauss-Koordinaten
     */   
    private long FValue_y(int x, int y, int t) {
        return (Y_GaussToPanel(f_im(X_PanelToGauss(x), Y_PanelToGauss(y))[t]));
    }

    /**
     * Diese Methode berechnet den Wert der Ableitungsfunktion
     *
     * @param x Real-Teil der komplexen Zahl in Panel-Koordinaten 
     * @param y Imaginaer-Teil der komplexen Zahl in Panel-Koordinaten
     * @param t Nummer des Graphen ("sqrt": 2, ansonsten 1)
     * @return zwei-elementiges Feld bestehend aus Real und Imaginaer-Teil der Ableitungsfunktion in Gauss-Koordinaten   
    */        
    private long[] DFValue(int x, int y, int t) {

        long[] ret = new long[2];
        ret[0] = X_GaussToPanel(df_re(X_PanelToGauss(x), Y_PanelToGauss(y))[t]);
        ret[1] = Y_GaussToPanel(df_im(X_PanelToGauss(x), Y_PanelToGauss(y))[t]);
        return ret;
    } 
  
    /**
     * Diese Methode berechnet die Approximation von f(z) nach dem Schema: f(z) = f(z0) + df(z0) * (z - z0)
     *
     * @param re Real-Teil der komplexen Zahl z0 in Panel-Koordinaten 
     * @param im Imaginaer-Teil der komplexen Zahl z0 in Panel-Koordinaten
     * @param rez Real-Teil der komplexen Zahl z in Panel-Koordinaten 
     * @param imz Imaginaer-Teil der komplexen Zahl z in Panel-Koordinaten
     * @param t Nummer des Graphen ("sqrt": 2, ansonsten 1)
     * @return zwei-elementiges Feld bestehend aus Real und Imaginaer-Teil der Approximation von f(z) in Gauss-Koordinaten   
    */        
    private long[] AFValue(int re, int im, int rez, int imz, int t) {
        double a,b,c,d,e,f,g,h;
        long[] df = DFValue(re, im, t);
        long[] ret = new long[2];

        a = X_PanelToGauss(re);
        b = Y_PanelToGauss(im); 
        c = X_PanelToGauss(rez);
        d = Y_PanelToGauss(imz);
        e = X_PanelToGauss(df[0]);
        f = Y_PanelToGauss(df[1]);
        g = X_PanelToGauss(FValue_x(re,im,t));
        h = Y_PanelToGauss(FValue_y(re,im,t));
        
        ret[0] = X_GaussToPanel( g + ( (c - a) * e - (d - b) * f ) );
        ret[1] = Y_GaussToPanel( h + ( (c - a) * f + (d - b) * e ) );

        return ret;
    }
    
    /**
     * Diese Methode rundet die Gauss-Werte des eingegebenen Panel-Wertes auf zwei Kommastellen 
     * Sie wird fuer die gaphische Ausgabe der Legende verwendet 
     *
     * @param x Real-Teil der komplexen Zahl in Panel-Koordinaten
     * @return Gerundeter Wert des Real-Teils in Gauss-Koordinaten
     */
    private double print_x(int x){
        return (Math.round(X_PanelToGauss(x)*100)/100.0);
    }

    /**
     * Diese Methode rundet die Gauss-Werte des eingegebenen Panel-Wertes auf zwei Kommastellen 
     * Sie wird fuer die gaphische Ausgabe der Legende verwendet 
     *
     * @param y Imaginaer-Teil der komplexen Zahl in Panel-Koordinaten
     * @return Gerundeter Wert des Imaginaer-Teils in Gauss-Koordinaten
     */
    private double print_y(int y){
        return -(Math.round(Y_PanelToGauss(y)*100)/100.0);
    }

    /**
     * Diese Methode dient der Berechnung der Real-Teile der Eckpunkte der Pfeilspitze
     *
     * @param start_re Real-Teil des Startvektors in Panel-Koordinaten
     * @param start_im Imaginaer-Teil des Startvektors in Panel-Koordinaten
     * @param end_re Real-Teil des Endvektors in Panel-Koordinaten
     * @param end_im Imaginaer-Teil des Endvektors in Panel-Koordinaten
     * @return drei-elementiges Feld, das die Real-Teile der Koordinaten der Eckpunkte der Pfeilspitze in Panel-Koordinaten beinhaltet    
     */
    private int[] getArrow_re (int start_re, int start_im, int end_re, int end_im){
        int[] ret = new int[3];
        double length, theta, gx, gy;

        gx = X_PanelToGauss(end_re);
        gy = Y_PanelToGauss(end_im);
        length = CartesianToPolar(gx,gy)[0] / scale;
        theta = CartesianToPolar(gx,gy)[1];
        
        if (length < 0) {
            theta += Math.PI;
            length *= -1;
        }

        ret[0] = (int) Math.ceil( start_re + length * Math.cos( -theta ) );
        ret[1] = (int) Math.ceil( ret[0] + 5 * Math.cos( -( theta + 5 * Math.PI / 4 ) ) );
        ret[2] = (int) Math.ceil( ret[0] + 5 * Math.cos( -( theta + 3 * Math.PI / 4 ) ) );

        return ret;
    }

    /**
     * Diese Methode dient der Berechnung der Imaginaer-Teile der Eckpunkte der Pfeilspitze
     *
     * @param start_re Real-Teil des Startvektors in Panel-Koordinaten
     * @param start_im Imaginaer-Teil des Startvektors in Panel-Koordinaten
     * @param end_re Real-Teil des Endvektors in Panel-Koordinaten
     * @param end_im Imaginaer-Teil des Endvektors in Panel-Koordinaten
     * @return drei-elementiges Feld, das die Imaginaer-Teile der Koordinaten der Eckpunkte der Pfeilspitze in Panel-Koordinaten beinhaltet    
     */
    private int[] getArrow_im (int start_re, int start_im, int end_re, int end_im){
        int[] ret = new int[3];
        double length, theta, gx, gy;

        gx = X_PanelToGauss(end_re);
        gy = Y_PanelToGauss(end_im);
        length = CartesianToPolar(gx,gy)[0] / scale;
        theta = CartesianToPolar(gx,gy)[1];

        if (length < 0) {
            theta += Math.PI;
            length *= -1;
        }

        ret[0] = (int) Math.ceil( start_im - length * Math.sin( -theta ) );
        ret[1] = (int) Math.ceil( ret[0] - 5 * Math.sin( -( theta + 5 * Math.PI / 4 ) ) );
        ret[2] = (int) Math.ceil( ret[0] - 5 * Math.sin( -( theta + 3 * Math.PI / 4 ) ) );

        return ret;
    }
    
    /**
     * Diese Methode dient der Definition der graphischen Ausgabe (Paint)
     * <br>Anmerkung: (cd) := Nur relevant, wenn die komplexe Ableitung aktiviert ist
     *
     * @param g Graphics-Objekt
     */
    public void paint(Graphics g) {
        /**
         * Diese Variablen werden fuer das Zeichnen des Eingabe- und Ausgabe-Rasters verwendet
         * bzw. fuer die Speicherung der zu berechnenden Werte der komplexen Ableitung (cd): cd_x, cd_y, rad := Radius der Punkte z0,z
         */  
        int x,y;
        int min_x = x1 - gridspace * gridlines; 
        int max_x = x1 + gridspace * gridlines;
        int min_y = y1 - gridspace * gridlines;
        int max_y = y1 + gridspace * gridlines;
        int x2[][] = new int[gridsize][gridsize]; 
        int y2[][] = new int[gridsize][gridsize]; 
        int cd_x[] = new int[7]; // (cd)
        int cd_y[] = new int[7]; // (cd)
        int unitlen = Math.round(1/scale);
        int rad = 8; // (cd)
        
        /** Diese Variablen definieren die Felder fuer die x und y Werte der Eckpunkte der Pfeilspitzen (cd) */        
        int[] arrow_x = new int[3];
        int[] arrow_y = new int[3];

        boolean infinity = false;
        
        /** Diese Variablen dienen der Speicherung der Werte fuer die graphische Ausgabe der Legende (cd) */        
        double print_x,print_y;

        // Werte des Ursprungs des Panel-Koordinatensystems
        origin_x = getSize().width/2;
        origin_y = getSize().height/2;

        // Arbeitsflaeche des Viewpanels
        parent.hiddengraphics.setColor(Color.white); 
        parent.hiddengraphics.fillRect(0,0,getSize().width,getSize().height); // Arbeitsflaeche leeren
        parent.hiddengraphics.setColor(Color.black);
        parent.hiddengraphics.drawRect(0,0,getSize().width-1,getSize().height-1); // aeussere Begrenzung
        parent.hiddengraphics.drawOval(getSize().width/2-unitlen,getSize().height/2-unitlen,unitlen*2,unitlen*2); // Mittelkreis
        parent.hiddengraphics.drawLine(getSize().width/2,0,getSize().width/2,getSize().height-1); // x-Achse
        parent.hiddengraphics.drawLine(0,getSize().height/2,getSize().width-1,getSize().height/2); // y-Achse
        parent.hiddengraphics.setFont(new Font("Monospaced",0,10));
        
        // Zeichnen der Ausgabe: 
        // 
        // 1. nicht (cd) 
        // 2. (cd)

        // 1. Eingabe Raster / Ausgabe Raster
        if ( !cd_state ) { 
            
            // Input Grid
            parent.hiddengraphics.setColor(Color.red);
            
            for (int i = -gridlines; i <= gridlines; i++) {
                x = x1 + i * gridspace; 
                parent.hiddengraphics.drawLine( x, min_y, x, max_y );
                y = y1 + i * gridspace;
                parent.hiddengraphics.drawLine( min_x, y, max_x, y );
            }
            
            // if log: infinity at 0 => infinity = true, if input-grid includes origin
            if( (this.function==Log) &&
                (signum(X_PanelToGauss(min_x)) != signum(X_PanelToGauss(max_x))) &&
                (signum(Y_PanelToGauss(min_y)) != signum(Y_PanelToGauss(max_y))) ) {
                infinity = true;
            }
            
            // Output Grid
            if (!infinity) { 
                for (int t = 0; t < CoG_perFunction[this.function]; t++) {
                    for (int i = 0; i < gridsize; i++) {
                        x = x1 + ( i - gridlines ) * gridspace; 
                        for (int j = 0; j < gridsize; j++) {
                            y = y1 + ( j - gridlines ) * gridspace;
                            x2[i][j] = (int) FValue_x(x, y, t);  
                            y2[i][j] = (int) FValue_y(x, y, t);  
                        }
                    }
                    
                    if(t==0)
                        parent.hiddengraphics.setColor(Color.blue);
                    else
                        parent.hiddengraphics.setColor(Color.green);
                    
                    // Painting Output Grid
                    for (int i = 0; i < gridsize; i++) {
                        for (int j = 1; j < gridsize; j++) {
                            parent.hiddengraphics.drawLine( x2[i][j-1], y2[i][j-1], x2[i][j], y2[i][j] );
                            parent.hiddengraphics.drawLine( x2[j-1][i], y2[j-1][i], x2[j][i], y2[j][i] );
                        }
                    }
                }
            }
            else
                parent.hiddengraphics.drawString("INFINITY",30,30);
        }
        // 2. Punkte, Vektoren, Linien, Legende (cd) 
        else { 
            
            // Variablen zur Positionierung des festen Textes
            int hlp1 = 30;
            int hlp2 = CoG_perFunction[this.function];
            
            // Ausgabe von z0 als Punkt mit dem Radius rad (s.o.) 
            cd_x[0] = init_z0_re; // Real-Teil von z0
            cd_y[0] = init_z0_im; // Imaginaer-Teil von z0
            parent.hiddengraphics.setColor(Color.black);
            parent.hiddengraphics.fillOval(cd_x[0]-rad/2,cd_y[0]-rad/2,rad,rad);
            parent.hiddengraphics.drawString("z0", cd_x[0]+rad/2, cd_y[0]-rad/2);
            // Legende
            print_x = print_x(cd_x[0]); 
            print_y = print_y(cd_y[0]);         
            parent.hiddengraphics.drawString("z0: " + "(" + print_x + " " + print_y + "i)", 30, hlp1);            
            
            
            for (int t = 0; t < hlp2; t++) {    
                
                // Ausgabe von F(z0) als Vektor (Pfeil) vom Nullpunkt 
                parent.hiddengraphics.setColor(Color.black);
                cd_x[2] = (int) FValue_x(cd_x[0],cd_y[0],t); // Real-Teil von F(z0)
                cd_y[2] = (int) FValue_y(cd_x[0],cd_y[0],t); // Imaginaer-Teil von F(z0)
                parent.hiddengraphics.drawLine(origin_x, origin_y, cd_x[2], cd_y[2]);
                // Berechnung der Werte fuer die Eckpunkte der Pfeilspitzen
                arrow_x = getArrow_re(origin_x, origin_y, cd_x[2], cd_y[2]);
                arrow_y = getArrow_im(origin_x, origin_y, cd_x[2], cd_y[2]);
                parent.hiddengraphics.fillPolygon(arrow_x, arrow_y, 3);
                // Legende
                print_x = print_x(cd_x[2]); 
                print_y = print_y(cd_y[2]); 
                if (t == 0) {
                    hlp1+=10;
                    parent.hiddengraphics.drawString("f(z0): " + "(" + print_x + " " + print_y + "i)", 30, hlp1);
                    if (shw1_state)
                        parent.hiddengraphics.drawString("f(z0)", cd_x[2]+rad/2, cd_y[2]-rad/2);
                }                
                else {
                    parent.hiddengraphics.setColor(Color.darkGray);
                    parent.hiddengraphics.drawString("(" + print_x + " " + print_y + "i)" , 71, 50);
                    if (shw1_state)
                        parent.hiddengraphics.drawString("f(z0)", cd_x[2]+rad/2, cd_y[2]-rad/2);
                }
                
                // Ausgabe von DF(z0) als Vektor (Pfeil) vom Nullpunkt 
                cd_x[3] = (int) DFValue(cd_x[0],cd_y[0],t)[0]; // Real-Teil von DF(z0)
                cd_y[3] = (int) DFValue(cd_x[0],cd_y[0],t)[1]; // Imaginaer-Teil DF(z0)
                parent.hiddengraphics.setColor(Color.green.darker());
                parent.hiddengraphics.drawLine(origin_x, origin_y, cd_x[3], cd_y[3]);
                // Berechnung der Werte fuer die Eckpunkte der Pfeilspitzen
                arrow_x = getArrow_re(origin_x, origin_y, cd_x[3], cd_y[3] );
                arrow_y = getArrow_im(origin_x, origin_y, cd_x[3], cd_y[3] );
                parent.hiddengraphics.fillPolygon(arrow_x, arrow_y, 3);
                // Legende
                print_x = print_x(cd_x[3]); 
                print_y = print_y(cd_x[3]); 
                if ((hlp2==2) && (t==0))
                    hlp1+=10;
                if (t == 0) {
                    hlp1+=10;
                    parent.hiddengraphics.drawString("df(z0): " + "(" + print_x + " " + print_y + "i)", 30, hlp1);
                    if (shw2_state)
                        parent.hiddengraphics.drawString("df(z0)", cd_x[3]+rad/2, cd_y[3]-rad/2);
                }                
                else {
                    parent.hiddengraphics.setColor(Color.green);
                    parent.hiddengraphics.drawString("(" + print_x + " " + print_y + "i)", 78, 70);   
                    if (shw2_state)
                        parent.hiddengraphics.drawString("df(z0)", cd_x[3]+rad/2, cd_y[3]-rad/2);
                }
                
            }
            
            
            // Ausgabe von z als Punkt mit dem Radius rad (s.o.) 
            cd_x[1] = init_z_re; // Real-Teil von z
            cd_y[1] = init_z_im; // Imaginaer-Teil von z
            parent.hiddengraphics.setColor(Color.blue.darker());
            parent.hiddengraphics.fillOval(cd_x[1]-rad/2,cd_y[1]-rad/2,rad,rad);
            parent.hiddengraphics.drawString("z ", cd_x[1]+rad/2, cd_y[1]-rad/2);
            // Legende
            print_x = print_x(cd_x[1]); 
            print_y = print_y(cd_y[1]);
            hlp1+=10;
            if (hlp2 == 2)
                hlp1+=10;
            parent.hiddengraphics.drawString("z: " + "(" + print_x + " " + print_y + "i)", 30, hlp1);
            

            for (int t = 0; t < hlp2; t++) {
                
                // Ausgabe von F(z) als Vektor (Pfeil) vom Nullpunkt 
                cd_x[5] = (int) FValue_x(cd_x[1],cd_y[1],t); // Real-Teil von F(z)
                cd_y[5] = (int) FValue_y(cd_x[1],cd_y[1],t); // Imaginaer-Teil von F(z)
                parent.hiddengraphics.setColor(Color.blue.darker());
                parent.hiddengraphics.drawLine(origin_x, origin_y, cd_x[5], cd_y[5]);
                // Berechnung der Werte fuer die Eckpunkte der Pfeilspitzen
                arrow_x = getArrow_re(origin_x, origin_y, cd_x[5], cd_y[5]);
                arrow_y = getArrow_im(origin_x, origin_y, cd_x[5], cd_y[5]);
                parent.hiddengraphics.fillPolygon(arrow_x, arrow_y, 3);
                // Legende
                print_x = print_x(cd_x[5]); 
                print_y = print_y(cd_y[5]); 
                if (t == 0) {
                    hlp1+=10;
                    parent.hiddengraphics.drawString("f(z): " + "(" + print_x + " " + print_y + "i)", 30, hlp1);
                    if (shw4_state) 
                        parent.hiddengraphics.drawString("f(z)", cd_x[5]+rad/2, cd_y[5]-rad/2);
                }
                else {
                    parent.hiddengraphics.setColor(Color.blue);
                    parent.hiddengraphics.drawString("(" + print_x + " " + print_y + "i)" , 66, 100);
                    if (shw4_state)
                        parent.hiddengraphics.drawString("f(z)", cd_x[5]+rad/2, cd_y[5]-rad/2);
                }
                
                // Ausgabe der Approximation* von F(z) als Vektor (Pfeil) vom Nullpunkt
                //
                // *: F(z0) + F'(z0) * (z - z0)         
                cd_x[4] = (int) AFValue(cd_x[0],cd_y[0],cd_x[1],cd_y[1],t)[0]; // Approximation des Real-Teils von F(z)
                cd_y[4] = (int) AFValue(cd_x[0],cd_y[0],cd_x[1],cd_y[1],t)[1]; // Approximation des Imaginaer-Teils von F(z)
                parent.hiddengraphics.setColor(Color.red.darker());
                parent.hiddengraphics.drawLine(origin_x, origin_y, cd_x[4], cd_y[4]);
                // Berechnung der Werte fuer die Eckpunkte der Pfeilspitzen             
                arrow_x = getArrow_re(origin_x, origin_y, cd_x[4], cd_y[4] );
                arrow_y = getArrow_im(origin_x, origin_y, cd_x[4], cd_y[4] );
                parent.hiddengraphics.fillPolygon(arrow_x, arrow_y, 3); 
                // Legende
                print_x = print_x(cd_x[4]); 
                print_y = print_y(cd_y[4]); 
                if ((hlp2 == 2) && (t == 0))
                    hlp1+=10;
                if (t==0) {
                    hlp1+=10;
                    parent.hiddengraphics.drawString("ap:f(z): " + "(" + print_x + " " + print_y + "i)", 30, hlp1);
                    if (shw3_state)
                        parent.hiddengraphics.drawString("f(z)apr", cd_x[4]+rad/2, cd_y[4]-rad/2);
                }
                else {
                    parent.hiddengraphics.setColor(Color.red);
                    parent.hiddengraphics.drawString("(" + print_x + " " + print_y + "i)", 84, 120);
                    if (shw3_state)                   
                        parent.hiddengraphics.drawString("ap:f(z)", cd_x[4]+rad/2, cd_y[4]-rad/2);
                }
                
            }

            // Ausgabe der Differenz (z - z0) als Verbindung vom Punkt z0 zum Punkt z
            parent.hiddengraphics.setColor(Color.blue.brighter());
            parent.hiddengraphics.drawLine(cd_x[0],cd_y[0],cd_x[1],cd_y[1]);
        }
        
        g.drawImage(parent.hiddenimagebuffer, 0,0, this);
    }
    
    public void update(Graphics g) {
        paint(g);
    }
    
    /** 
     * Diese Methode beschreibt die Aktion bei Betaetigung des Maus-Buttons 
     * <br> - nicht (cd): kann die neue Position des Eingabe Rasters festgelegt werden (ziehen mglich)
     * <br> - (cd): kann die neue Position von z festgelegt werden (ziehen mglich)
     * <br> - (cd) && "choose z0": kann die neue Position von z0 festgelegt werden (ziehen nicht mglich)
    */    
    public boolean mouseDown(java.awt.Event evt, int x, int y) {
        if (contains(x,y)) {
            if ((cd_state) && (z0_state)) {
                init_z0_re = x;
                init_z0_im = y;
                z0_state = false;
            }
            else if ((cd_state) && (!z0_state)) {
                init_z_re = x;
                init_z_im = y;
            }
            else {
                x1 = x;
                y1 = y; 
            }
            
            repaint();
        }
        return true;
    }

    /** 
     * Diese Methode beschreibt die Aktion bei Ziehen der Maus bei gedruecktem Maus-Button 
     * <br> - nicht (cd): kann die Position des Eingabe Rasters durch ziehen veraendert werden
     * <br> - (cd): kann die Position von z durch ziehen veraendert werden
     */      
    public boolean mouseDrag(java.awt.Event evt, int x, int y) {
        if (contains(x,y)) {
            if (cd_state) {
                init_z_re = x;
                init_z_im = y;
            }
            else {
                x1 = x;
                y1 = y;
            }

            repaint();
        }
        return true;
    }
}


/**
 * Die Klasse enthaelt die Controls fuer das Applet
 * <br> - Funktionsauswahl
 * <br> - Checkbox fuer das Aktivieren der komplexen Ableitung
 * <br> - Button fuer das Neu-Setzen von z0
 * <br> - 4 Checkboxen fuer die Anzeige der Beschriftung von f(z0), df(z0), ap:f(z), f(z)
 */
class Controls extends Panel {
    
    ViewPanel target;
    List lst;
    Checkbox cd;
    Checkbox shw1,shw2,shw3,shw4;
    Button setz0;

    Panel zoompanel = new Panel();
    Panel cdpanel = new Panel();

    Scrollbar scrollbar;

    /**
     * Konstruktor fuer die Klasse Controls
     *
     * @param target Instanz der Klasse ViewPanel, die gesteuert werden soll
     * @param size Groesse des Viewpanels, das gesteuert werden soll
     */    
    public Controls (ViewPanel target, Dimension size) {

        this.target = target;
        setLayout(new BorderLayout());
        setBackground(new Color(155,155,245));

        lst = new List(8);
        for(int i = 0; i< ViewPanel.function_names.length; i++)
            lst.add(ViewPanel.function_names[i], i);
        this.add(lst, BorderLayout.NORTH);
        lst.addItemListener(new ItemListener() {
                public void itemStateChanged(ItemEvent ie) {
                    ItemEventHandler(ie);
                }
            });
        
        cdpanel.setLayout(new GridLayout(6,1));
        cd = new Checkbox(" derivative");
        cd.setState(false);
        cd.addItemListener(new ItemListener(){
                public void itemStateChanged(ItemEvent ie) {
                    CDEventHandler(ie);
                }
            });
        setz0 = new Button(" choose  z0");
        setz0.addActionListener(new ActionListener(){
                public void actionPerformed(ActionEvent ae) {
                    Z0EventHandler(ae);
                }
            });

        shw1 = new Checkbox(" show f(z0)");
        shw1.setState(false);
        shw1.addItemListener(new ItemListener(){
                public void itemStateChanged(ItemEvent ie) {
                    SHWEventHandler(ie);
                }
            }); 
        shw2 = new Checkbox(" show df(z0)");
        shw2.setState(false);
        shw2.addItemListener(new ItemListener(){
                public void itemStateChanged(ItemEvent ie) {
                    SHWEventHandler(ie);
                }
            }); 
        shw3 = new Checkbox(" show ap:f(z)");
        shw3.setState(false);
        shw3.addItemListener(new ItemListener(){
                public void itemStateChanged(ItemEvent ie) {
                    SHWEventHandler(ie);
                }
            }); 
        shw4 = new Checkbox(" show f(z)");
        shw4.setState(false);
        shw4.addItemListener(new ItemListener(){
                public void itemStateChanged(ItemEvent ie) {
                    SHWEventHandler(ie);
                }
            }); 
        
        cdpanel.add(cd);
        cdpanel.add(setz0);        
        cdpanel.add(shw1);
        cdpanel.add(shw2);
        cdpanel.add(shw3);
        cdpanel.add(shw4);
        this.add(cdpanel, BorderLayout.CENTER);

        zoompanel.setLayout(new GridLayout(2,1));
        zoompanel.add(new Label("Zoomen:"));
        scrollbar = new Scrollbar(Scrollbar.HORIZONTAL,50,1,5,100);
        scrollbar.addAdjustmentListener(new AdjustmentListener(){
                public void adjustmentValueChanged(AdjustmentEvent e)  {
                    AdjustmentEventHandler(e);
                }
            });
        zoompanel.add(scrollbar);
        this.add(zoompanel,BorderLayout.SOUTH);
    }

    /**
     * Diese Methode definiert die Aktion bei Eintreten einer Funktionsauswahl
     */
    private void ItemEventHandler(ItemEvent ie) {
        if (ie.getSource() instanceof List) {
            target.setFunction(((List)ie.getSource()).getSelectedIndex());
            target.repaint();
        }
    }
    
    /**
     * Diese Methode definiert die Aktion bei Veraenderung des Zoom Scrollbars
     */    
    private void AdjustmentEventHandler(AdjustmentEvent ae) {
        if(ae.getSource() instanceof Scrollbar) {
            target.scale = 0.001f*scrollbar.getValue();
            target.repaint();
        }
    }

    /**
     * Diese Methode definiert die Aktion bei Veraenderung des Status der Checkbox "derivative"
     */ 
    private void CDEventHandler(ItemEvent ie) {
        if (cd.getState()) {
            target.cd_state=true;
            target.init_z0_re = target.getSize().width/2;
            target.init_z0_im = target.getSize().height/2;
            target.init_z_re = target.getSize().width/2 + 50;
            target.init_z_im = target.getSize().height/2;
            target.repaint();
        }
        else {
            target.cd_state=false;
            target.z0_state=false;
            target.repaint();
        }
    }

    /**
     * Diese Methode definiert die Aktion bei Veraenderung des Status der Checkboxen fuer die graphische Bezeichnung der einzelen Elemente (cd)
     */
    private void SHWEventHandler(ItemEvent ie) {

        if (shw1.getState()) {
            target.shw1_state=true;
            target.repaint();
        }
        else {
            target.shw1_state=false;
            target.repaint();
        }
        if (shw2.getState()) {
            target.shw2_state=true;
            target.repaint();
        }
        else {
            target.shw2_state=false;
            target.repaint();
        }
        if (shw3.getState()) {
            target.shw3_state=true;
            target.repaint();
        }
        else {
            target.shw3_state=false;
            target.repaint();
        }
        if (shw4.getState()) {
            target.shw4_state=true;
            target.repaint();
        }
        else {
            target.shw4_state=false;
            target.repaint();
        }

    }
    
    /**
     * Diese Methode definiert die Aktion bei Anwendung des "choose z0" Buttons (cd)
     */    
    private void Z0EventHandler(ActionEvent ae) {
        target.z0_state=true;
    }
}

/** 
 * EOF
 */  

    
